@@ -4,10 +4,18 @@ import logging
 import operator
 import os
 import re
-from collections import OrderedDict
 from tempfile import NamedTemporaryFile
+from typing import Callable, Dict, List, Optional, Union
 
-import galaxy.model
+from sqlalchemy.orm.scoping import ScopedSession
+
+from galaxy.model import (
+    HistoryDatasetAssociation,
+    HistoryDatasetCollectionAssociation,
+    Job,
+    JobToOutputDatasetAssociation,
+    LibraryDatasetDatasetAssociation,
+)
 from galaxy.model.dataset_collections import builder
 from galaxy.model.dataset_collections.structure import UninitializedTree
 from galaxy.model.dataset_collections.type_description import COLLECTION_TYPE_DESCRIPTION_FACTORY
@@ -15,15 +23,17 @@ from galaxy.model.store.discover import (
     discover_target_directory,
     DiscoveredFile,
     JsonCollectedDatasetMatch,
+    MetadataSourceProvider as AbstractMetadataSourceProvider,
     ModelPersistenceContext,
+    PermissionProvider as AbstractPermissionProvider,
     persist_elements_to_folder,
     persist_elements_to_hdca,
-    persist_extra_files,
     persist_hdas,
     RegexCollectedDatasetMatch,
     SessionlessModelPersistenceContext,
     UNSET,
 )
+from galaxy.objectstore import ObjectStore
 from galaxy.tool_util.parser.output_collection_def import (
     DEFAULT_DATASET_COLLECTOR_DESCRIPTION,
     INPUT_DBKEY_TOKEN,
@@ -34,7 +44,8 @@ from galaxy.tool_util.parser.output_objects import (
     ToolOutputCollection,
 )
 from galaxy.util import (
-    unicodify
+    shrink_and_unicodify,
+    unicodify,
 )
 
 DATASET_ID_TOKEN = "DATASET_ID"
@@ -44,7 +55,7 @@ log = logging.getLogger(__name__)
 
 # PermissionProvider and MetadataSourceProvider are abstractions over input data used to
 # collect and produce dynamic outputs.
-class PermissionProvider:
+class PermissionProvider(AbstractPermissionProvider):
 
     def __init__(self, inp_data, security_agent, job):
         self._job = job
@@ -76,7 +87,7 @@ class PermissionProvider:
         self._security_agent.copy_dataset_permissions(init_from.dataset, primary_data.dataset)
 
 
-class MetadataSourceProvider:
+class MetadataSourceProvider(AbstractMetadataSourceProvider):
 
     def __init__(self, inp_data):
         self._inp_data = inp_data
@@ -105,9 +116,10 @@ def collect_dynamic_outputs(
         # "hdca" (place discovered files in a history dataset collection), and "hdas" (place discovered files in a history
         # as stand-alone datasets).
         if destination_type == "library_folder":
-            # populate a library folder (needs to be already have been created)
+            # populate a library folder (needs to have already been created)
             library_folder = job_context.get_library_folder(destination)
             persist_elements_to_folder(job_context, elements, library_folder)
+            job_context.persist_library_folder(library_folder)
         elif destination_type == "hdca":
             # create or populate a dataset collection in the history
             assert "collection_type" in unnamed_output_dict
@@ -120,7 +132,13 @@ def collect_dynamic_outputs(
                 collection_type_description = COLLECTION_TYPE_DESCRIPTION_FACTORY.for_collection_type(collection_type)
                 structure = UninitializedTree(collection_type_description)
                 hdca = job_context.create_hdca(name, structure)
-            persist_elements_to_hdca(job_context, elements, hdca, collector=DEFAULT_DATASET_COLLECTOR)
+                output_collections[name] = hdca
+                job_context.add_dataset_collection(hdca)
+            error_message = unnamed_output_dict.get("error_message")
+            if error_message:
+                hdca.collection.handle_population_failed(error_message)
+            else:
+                persist_elements_to_hdca(job_context, elements, hdca, collector=DEFAULT_DATASET_COLLECTOR)
         elif destination_type == "hdas":
             persist_hdas(elements, job_context, final_job_state=job_context.final_job_state)
 
@@ -165,36 +183,93 @@ def collect_dynamic_outputs(
 
 class BaseJobContext:
 
+    max_discovered_files: Union[int, float]
+
     def add_dataset_collection(self, collection):
         pass
 
     def find_files(self, output_name, collection, dataset_collectors):
-        filenames = OrderedDict()
+        filenames = {}
         for discovered_file in discover_files(output_name, self.tool_provided_metadata, dataset_collectors, self.job_working_directory, collection):
+            self.increment_discovered_file_count()
             filenames[discovered_file.path] = discovered_file
         return filenames
 
+    def get_job_id(self):
+        return None  # overwritten in subclasses
+
 
 class JobContext(ModelPersistenceContext, BaseJobContext):
 
-    def __init__(self, tool, tool_provided_metadata, job, job_working_directory, permission_provider, metadata_source_provider, input_dbkey, object_store, final_job_state):
+    def __init__(
+            self,
+            tool,
+            tool_provided_metadata,
+            job,
+            job_working_directory,
+            permission_provider,
+            metadata_source_provider,
+            input_dbkey,
+            object_store,
+            final_job_state,
+            max_discovered_files: Optional[int],
+            flush_per_n_datasets=None,
+    ):
         self.tool = tool
-        self.metadata_source_provider = metadata_source_provider
-        self.permission_provider = permission_provider
-        self.input_dbkey = input_dbkey
+        self._metadata_source_provider = metadata_source_provider
+        self._permission_provider = permission_provider
+        self._input_dbkey = input_dbkey
         self.app = tool.app
-        self.sa_session = tool.sa_session
-        self.job = job
+        self._sa_session = tool.sa_session
+        self._job = job
         self.job_working_directory = job_working_directory
         self.tool_provided_metadata = tool_provided_metadata
-        self.object_store = object_store
+        self._object_store = object_store
         self.final_job_state = final_job_state
+        self._flush_per_n_datasets = flush_per_n_datasets
+        self.max_discovered_files = float('inf') if max_discovered_files is None else max_discovered_files
+        self.discovered_file_count = 0
+        self._tag_handler = None
+
+    @property
+    def tag_handler(self):
+        if self._tag_handler is None:
+            self._tag_handler = self.app.tag_handler.create_tag_handler_session()
+        return self._tag_handler
 
     @property
     def work_context(self):
         from galaxy.work.context import WorkRequestContext
         return WorkRequestContext(self.app, user=self.user)
 
+    @property
+    def sa_session(self) -> ScopedSession:
+        return self._sa_session
+
+    @property
+    def permission_provider(self) -> PermissionProvider:
+        return self._permission_provider
+
+    @property
+    def metadata_source_provider(self) -> MetadataSourceProvider:
+        return self._metadata_source_provider
+
+    @property
+    def job(self) -> Job:
+        return self._job
+
+    @property
+    def flush_per_n_datasets(self) -> Optional[int]:
+        return self._flush_per_n_datasets
+
+    @property
+    def input_dbkey(self) -> str:
+        return self._input_dbkey
+
+    @property
+    def object_store(self) -> ObjectStore:
+        return self._object_store
+
     @property
     def user(self):
         if self.job:
@@ -203,10 +278,6 @@ class JobContext(ModelPersistenceContext, BaseJobContext):
             user = None
         return user
 
-    @property
-    def tag_handler(self):
-        return self.app.tag_handler
-
     def persist_object(self, obj):
         self.sa_session.add(obj)
 
@@ -220,7 +291,7 @@ class JobContext(ModelPersistenceContext, BaseJobContext):
         return library_folder
 
     def get_hdca(self, object_id):
-        hdca = self.sa_session.query(galaxy.model.HistoryDatasetCollectionAssociation).get(int(object_id))
+        hdca = self.sa_session.query(HistoryDatasetCollectionAssociation).get(int(object_id))
         return hdca
 
     def create_library_folder(self, parent_folder, name, description):
@@ -232,14 +303,14 @@ class JobContext(ModelPersistenceContext, BaseJobContext):
     def create_hdca(self, name, structure):
         history = self.job.history
         trans = self.work_context
-        collections_service = self.app.dataset_collections_service
-        hdca = collections_service.precreate_dataset_collection_instance(
+        collection_manager = self.app.dataset_collection_manager
+        hdca = collection_manager.precreate_dataset_collection_instance(
             trans, history, name, structure=structure
         )
         return hdca
 
     def add_output_dataset_association(self, name, dataset):
-        assoc = galaxy.model.JobToOutputDatasetAssociation(name, dataset)
+        assoc = JobToOutputDatasetAssociation(name, dataset)
         assoc.job = self.job
         self.sa_session.add(assoc)
 
@@ -266,7 +337,8 @@ class JobContext(ModelPersistenceContext, BaseJobContext):
 
     def add_datasets_to_history(self, datasets, for_output_dataset=None):
         sa_session = self.sa_session
-        self.job.history.add_datasets(sa_session, datasets)
+        self.job.history.stage_addition(datasets)
+        pending_histories = {self.job.history}
         if for_output_dataset is not None:
             # Need to update all associated output hdas, i.e. history was
             # shared with job running
@@ -275,9 +347,11 @@ class JobContext(ModelPersistenceContext, BaseJobContext):
                     continue
                 for dataset in datasets:
                     new_data = dataset.copy()
-                    copied_dataset.history.add_dataset(new_data)
+                    copied_dataset.history.stage_addition(new_data)
+                    pending_histories.add(copied_dataset.history)
                     sa_session.add(new_data)
-                    sa_session.flush()
+        for history in pending_histories:
+            history.add_pending_items()
 
     def output_collection_def(self, name):
         tool = self.tool
@@ -296,19 +370,24 @@ class JobContext(ModelPersistenceContext, BaseJobContext):
     def job_id(self):
         return self.job.id
 
+    def get_job_id(self):
+        return self.job.id
+
+    def get_implicit_collection_jobs_association_id(self):
+        return self.job.implicit_collection_jobs_association and self.job.implicit_collection_jobs_association.id
+
 
 class SessionlessJobContext(SessionlessModelPersistenceContext, BaseJobContext):
 
-    def __init__(self, metadata_params, tool_provided_metadata, object_store, export_store, import_store, working_directory, final_job_state):
+    def __init__(self, metadata_params, tool_provided_metadata, object_store, export_store, import_store, working_directory, final_job_state, max_discovered_files: Optional[int]):
         # TODO: use a metadata source provider... (pop from inputs and add parameter)
-        # TODO: handle input_dbkey...
-        input_dbkey = "?"
         super().__init__(object_store, export_store, working_directory)
         self.metadata_params = metadata_params
         self.tool_provided_metadata = tool_provided_metadata
         self.import_store = import_store
-        self.input_dbkey = input_dbkey
         self.final_job_state = final_job_state
+        self.max_discovered_files = float('inf') if max_discovered_files is None else max_discovered_files
+        self.discovered_file_count = 0
 
     def output_collection_def(self, name):
         tool_as_dict = self.metadata_params["tool"]
@@ -334,7 +413,7 @@ class SessionlessJobContext(SessionlessModelPersistenceContext, BaseJobContext):
         return "non-session bound job"
 
     def get_hdca(self, object_id):
-        hdca = self.import_store.sa_session.query(galaxy.model.HistoryDatasetCollectionAssociation).find(int(object_id))
+        hdca = self.import_store.sa_session.query(HistoryDatasetCollectionAssociation).find(int(object_id))
         if hdca:
             self.export_store.add_dataset_collection(hdca)
             for collection_dataset in hdca.dataset_instances:
@@ -352,32 +431,39 @@ class SessionlessJobContext(SessionlessModelPersistenceContext, BaseJobContext):
             self.export_store.collection_datasets[collection_dataset.id] = True
 
     def add_output_dataset_association(self, name, dataset_instance):
-        job_id = self.metadata_params["job_id_tag"]
-        self.export_store.add_job_output_dataset_associations(job_id, name, dataset_instance)
+        self.export_store.add_job_output_dataset_associations(self.get_job_id(), name, dataset_instance)
+
+    def get_job_id(self):
+        return self.metadata_params["job_id_tag"]
+
+    def get_implicit_collection_jobs_association_id(self):
+        return self.metadata_params.get("implicit_collection_jobs_association_id")
 
 
-def collect_primary_datasets(job_context, output, input_ext):
+def collect_primary_datasets(job_context: Union[JobContext, SessionlessJobContext], output, input_ext):
     job_working_directory = job_context.job_working_directory
 
     # Loop through output file names, looking for generated primary
     # datasets in form specified by discover dataset patterns or in tool provided metadata.
     primary_output_assigned = False
     new_outdata_name = None
-    primary_datasets = {}
+    primary_datasets: Dict[str, Dict[str, Union[HistoryDatasetAssociation, LibraryDatasetDatasetAssociation]]] = {}
+    storage_callbacks: List[Callable] = []
     for output_index, (name, outdata) in enumerate(output.items()):
         dataset_collectors = [DEFAULT_DATASET_COLLECTOR]
         output_def = job_context.output_def(name)
         if output_def is not None:
             dataset_collectors = [dataset_collector(description) for description in output_def.dataset_collector_descriptions]
-        filenames = OrderedDict()
+        filenames = {}
         for discovered_file in discover_files(name, job_context.tool_provided_metadata, dataset_collectors, job_working_directory, outdata):
+            job_context.increment_discovered_file_count()
             filenames[discovered_file.path] = discovered_file
         for filename_index, (filename, discovered_file) in enumerate(filenames.items()):
             extra_file_collector = discovered_file.collector
             fields_match = discovered_file.match
             if not fields_match:
                 # Before I guess pop() would just have thrown an IndexError
-                raise Exception("Problem parsing metadata fields for file %s" % filename)
+                raise Exception(f"Problem parsing metadata fields for file {filename}")
             designation = fields_match.designation
             ext = fields_match.ext
             if ext == "input":
@@ -386,7 +472,7 @@ def collect_primary_datasets(job_context, output, input_ext):
             if dbkey == INPUT_DBKEY_TOKEN:
                 dbkey = job_context.input_dbkey
             if filename_index == 0 and extra_file_collector.assign_primary_output and output_index == 0:
-                new_outdata_name = fields_match.name or "{} ({})".format(outdata.name, designation)
+                new_outdata_name = fields_match.name or f"{outdata.name} ({designation})"
                 outdata.change_datatype(ext)
                 outdata.dbkey = dbkey
                 outdata.designation = designation
@@ -396,16 +482,20 @@ def collect_primary_datasets(job_context, output, input_ext):
                 primary_output_assigned = True
                 continue
             if name not in primary_datasets:
-                primary_datasets[name] = OrderedDict()
+                primary_datasets[name] = {}
             visible = fields_match.visible
             # Create new primary dataset
-            new_primary_name = fields_match.name or "{} ({})".format(outdata.name, designation)
+            new_primary_name = fields_match.name or f"{outdata.name} ({designation})"
             info = outdata.info
 
             # TODO: should be able to disambiguate files in different directories...
             new_primary_filename = os.path.split(filename)[-1]
             new_primary_datasets_attributes = job_context.tool_provided_metadata.get_new_dataset_meta_by_basename(name, new_primary_filename)
-
+            extra_files = None
+            if new_primary_datasets_attributes:
+                extra_files_path = new_primary_datasets_attributes.get('extra_files', None)
+                if extra_files_path:
+                    extra_files = os.path.join(job_working_directory, extra_files_path)
             primary_data = job_context.create_dataset(
                 ext,
                 designation,
@@ -413,18 +503,15 @@ def collect_primary_datasets(job_context, output, input_ext):
                 dbkey,
                 new_primary_name,
                 filename,
+                extra_files=extra_files,
                 info=info,
                 init_from=outdata,
                 dataset_attributes=new_primary_datasets_attributes,
+                creating_job_id=job_context.get_job_id() if job_context else None,
+                storage_callbacks=storage_callbacks
             )
             # Associate new dataset with job
-            job_context.add_output_dataset_association('__new_primary_file_{}|{}__'.format(name, designation), primary_data)
-
-            if new_primary_datasets_attributes:
-                extra_files_path = new_primary_datasets_attributes.get('extra_files', None)
-                if extra_files_path:
-                    extra_files_path_joined = os.path.join(job_working_directory, extra_files_path)
-                    persist_extra_files(job_context.object_store, extra_files_path_joined, primary_data)
+            job_context.add_output_dataset_association(f'__new_primary_file_{name}|{designation}__', primary_data)
             job_context.add_datasets_to_history([primary_data], for_output_dataset=outdata)
             # Add dataset to return dict
             primary_datasets[name][designation] = primary_data
@@ -437,7 +524,9 @@ def collect_primary_datasets(job_context, output, input_ext):
             if sa_session:
                 sa_session.add(outdata)
 
-    job_context.flush()
+    # Move discovered outputs to storage and set metdata / peeks
+    for callback in storage_callbacks:
+        callback()
     return primary_datasets
 
 
@@ -464,27 +553,35 @@ def walk_over_file_collectors(extra_file_collectors, job_working_directory, matc
             yield match, extra_file_collector
 
 
-def walk_over_extra_files(target_dir, extra_file_collector, job_working_directory, matchable):
+def walk_over_extra_files(target_dir, extra_file_collector, job_working_directory, matchable, parent_paths=None):
     """
     Walks through all files in a given directory, and returns all files that
     match the given collector's match criteria. If the collector has the
     recurse flag enabled, will also recursively descend into child folders.
     """
-    matches = []
-    directory = discover_target_directory(target_dir, job_working_directory)
-    if os.path.isdir(directory):
-        for filename in os.listdir(directory):
-            path = os.path.join(directory, filename)
-            if os.path.isdir(path) and extra_file_collector.recurse:
-                # The current directory is already validated, so use that as the next job_working_directory when recursing
-                for match in walk_over_extra_files(filename, extra_file_collector, directory, matchable):
-                    yield match
-            else:
-                match = extra_file_collector.match(matchable, filename, path=path)
-                if match:
-                    matches.append(match)
+    parent_paths = parent_paths or []
 
-    for match in extra_file_collector.sort(matches):
+    def _walk(target_dir, extra_file_collector, job_working_directory, matchable, parent_paths):
+        directory = discover_target_directory(target_dir, job_working_directory)
+        if os.path.isdir(directory):
+            for filename in os.listdir(directory):
+                path = os.path.join(directory, filename)
+                if os.path.isdir(path):
+                    if extra_file_collector.recurse:
+                        new_parent_paths = parent_paths[:]
+                        new_parent_paths.append(filename)
+                        # The current directory is already validated, so use that as the next job_working_directory when recursing
+                        yield from _walk(
+                            filename, extra_file_collector, directory, matchable, parent_paths=new_parent_paths
+                        )
+                else:
+                    match = extra_file_collector.match(matchable, filename, path=path, parent_paths=parent_paths)
+                    if match:
+                        yield match
+
+    for match in extra_file_collector.sort(
+        _walk(target_dir, extra_file_collector, job_working_directory, matchable, parent_paths)
+    ):
         yield match
 
 
@@ -527,6 +624,7 @@ class DatasetCollector:
         self.directory = dataset_collection_description.directory
         self.assign_primary_output = dataset_collection_description.assign_primary_output
         self.recurse = dataset_collection_description.recurse
+        self.match_relative_path = dataset_collection_description.match_relative_path
 
     def _pattern_for_dataset(self, dataset_instance=None):
         token_replacement = r'\d+'
@@ -534,8 +632,10 @@ class DatasetCollector:
             token_replacement = str(dataset_instance.id)
         return self.pattern.replace(DATASET_ID_TOKEN, token_replacement)
 
-    def match(self, dataset_instance, filename, path=None):
+    def match(self, dataset_instance, filename, path=None, parent_paths=None):
         pattern = self._pattern_for_dataset(dataset_instance)
+        if self.match_relative_path and parent_paths:
+            filename = os.path.join(*parent_paths, filename)
         re_match = re.match(pattern, filename)
         match_object = None
         if re_match:
@@ -577,14 +677,14 @@ def read_exit_code_from(exit_code_file, id_tag):
         exit_code = int(exit_code_str)
     except ValueError:
         galaxy_id_tag = id_tag
-        log.warning("({}) Exit code '{}' invalid. Using 0.".format(galaxy_id_tag, exit_code_str))
+        log.warning(f"({galaxy_id_tag}) Exit code '{exit_code_str}' invalid. Using 0.")
         exit_code = 0
 
     return exit_code
 
 
 def default_exit_code_file(files_dir, id_tag):
-    return os.path.join(files_dir, 'galaxy_%s.ec' % id_tag)
+    return os.path.join(files_dir, f'galaxy_{id_tag}.ec')
 
 
 def collect_extra_files(object_store, dataset, job_working_directory):
@@ -596,7 +696,7 @@ def collect_extra_files(object_store, dataset, job_working_directory):
         # automatically creates them.  However, empty directories will
         # not be created in the object store at all, which might be a
         # problem.
-        for root, dirs, files in os.walk(temp_file_path):
+        for root, _dirs, files in os.walk(temp_file_path):
             extra_dir = root.replace(os.path.join(job_working_directory, "working"), '', 1).lstrip(os.path.sep)
             for f in files:
                 object_store.update_from_file(
@@ -620,3 +720,11 @@ def collect_extra_files(object_store, dataset, job_working_directory):
                 dataset.set_size()
         except Exception as e:
             log.warning('Unable to generate primary composite file automatically for %s: %s', dataset.dataset.id, unicodify(e))
+
+
+def collect_shrinked_content_from_path(path):
+    try:
+        with open(path, 'rb') as fh:
+            return shrink_and_unicodify(fh.read().strip())
+    except FileNotFoundError:
+        return None
